import yaml

from cfnjsontoyaml.mixins.type_checker import TypeChecker


class ImportValue(yaml.YAMLObject, TypeChecker):
    yaml_tag = u'!ImportValue'

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return '{tag}({value})'.format(tag=self.__class__.__name__, value=self.value)

    @staticmethod
    def representer(dumper, data):
        return dumper.represent_scalar(u'!ImportValue', data.value, style='double-quote')

    @staticmethod
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return ImportValue(value)

if __name__ == 'cfnjsontoyaml.yamlobject.importvalue':
    yaml.add_representer(ImportValue, ImportValue.representer)
    yaml.add_constructor(u'!ImportValue', ImportValue.constructor)



class TypeChecker(object):
    FUNCTIONS = [
        'Ref',
        'Fn::Base64',
        'Fn::FindInMap',
        'Fn::GetAtt',
        'Fn::GetAZs',
        'Fn::ImportValue',
        'Fn::Join',
        'Fn::Select',
        'Fn::Sub',
        'Fn::And',
        'Fn::Equals',
        'Fn::If',
        'Fn::Not',
        'Fn::Or'
    ]

    def _is_function(self, value):
        return (
            (not self._is_array(value)) and
            (len(value) == 1) and
            (self._get_function_key(value) in self.FUNCTIONS)
        )

    def _get_function_key(self, value):
        return value.keys()[0]

    def _is_array(self, value):
        return isinstance(value, (list, tuple))



import yaml

from cfnjsontoyaml.yamlobject.join import Join


class Base64(yaml.YAMLObject):
    yaml_tag = u'!Base64'

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return '{tag}({value})'.format(tag=self.__class__.__name__, value=self.value)

    @staticmethod
    def representer(dumper, data):
        if type(data.value) in [str, unicode, Join]:
            return dumper.represent_scalar(u'!Base64', data.value, style='double-quoted')
        else:
            return dumper.represent_sequence(u'!Base64', data.value, flow_style=True)

    @staticmethod
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return Base64(value)

if __name__ == 'cfnjsontoyaml.yamlobject.base64':
    yaml.add_representer(Base64, Base64.representer)
    yaml.add_constructor(u'!Base64', Base64.constructor)



from yaml.dumper import Dumper
from yaml.representer import SafeRepresenter

class CleanDumper(Dumper):
   pass

CleanDumper.add_representer(str,
       SafeRepresenter.represent_str)

CleanDumper.add_representer(unicode,
        SafeRepresenter.represent_unicode)



from cfnjsontoyaml.yamlobject.base64 import Base64
from cfnjsontoyaml.yamlobject.equals import Equals
from cfnjsontoyaml.yamlobject.findinmap import FindInMap
from cfnjsontoyaml.yamlobject.fnand import And
from cfnjsontoyaml.yamlobject.fnif import If
from cfnjsontoyaml.yamlobject.fnnot import Not
from cfnjsontoyaml.yamlobject.fnor import Or
from cfnjsontoyaml.yamlobject.getatt import GetAtt
from cfnjsontoyaml.yamlobject.getazs import GetAZs
from cfnjsontoyaml.yamlobject.importvalue import ImportValue
from cfnjsontoyaml.yamlobject.join import Join
from cfnjsontoyaml.yamlobject.ref import Ref
from cfnjsontoyaml.yamlobject.select import Select
from cfnjsontoyaml.yamlobject.sub import Sub

FUNCTION_MAPPING = {
    'Ref': Ref,
    'Fn::Base64': Base64,
    'Fn::FindInMap': FindInMap,
    'Fn::GetAtt': GetAtt,
    'Fn::GetAZs':GetAZs,
    'Fn::ImportValue': ImportValue,
    'Fn::Join': Join,
    'Fn::Select': Select,
    'Fn::Sub': Sub,
    'Fn::And': And,
    'Fn::Equals': Equals,
    'Fn::If': If,
    'Fn::Not': Not,
    'Fn::Or': Or
}



import yaml

from cfnjsontoyaml.parser.subbuilder import SubBuilder


class Sub(yaml.YAMLObject):
    yaml_tag = u'!Sub'

    def __init__(self, value, use_literal=False):
        pattern, substitutions = value
        if substitutions is None:
            self.value = pattern
        else:
            self.value = value

        self.use_literal = use_literal

    def __repr__(self):
        return '{tag}({value})'.format(tag=self.__class__.__name__, value=self.value)

    @property
    def is_sequence(self):
        return type(self.value) in [list, tuple]

    @staticmethod
    def representer(dumper, data):
        if data.is_sequence:

            use_literal = "\n" in data.value[0]
            if use_literal:
                pattern = literal(data.value[0])
            else:
                pattern = data.value[0]

            substitutions = data.value[1]

            return dumper.represent_sequence(u'!Sub', [pattern, substitutions], flow_style=False)
        else:
            return dumper.represent_scalar(u'!Sub', data.value, style=('|' if data.use_literal else '"'))

    @staticmethod
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return Sub(value)


class literal(str): pass


def literal_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')


if __name__ == 'cfnjsontoyaml.yamlobject.sub':
    yaml.add_representer(Sub, Sub.representer)
    yaml.add_constructor(u'!Ref', Sub.constructor)
    yaml.add_representer(literal, literal_presenter)



from cfnjsontoyaml.mixins.type_checker import TypeChecker
from cfnjsontoyaml.yamlobject.importvalue import ImportValue
from cfnjsontoyaml.yamlobject.join import Join
from cfnjsontoyaml.yamlobject.ref import Ref
from cfnjsontoyaml.yamlobject.sub import Sub


class NodeParser(object):
    PARSER_OPTIONS = {
        'join_to_sub': False,
        'use_tag': True
    }

    def __init__(self, node, **options):

        # Set option defaults
        self.options = self.PARSER_OPTIONS.copy()
        self.options.update(options)

        if type(node) == dict:
            self._nodes = map(lambda node_pair: Node(*node_pair, parser=self, parent=None), node.items())
            self._node_type = dict

    def parse(self):

        for node in self._nodes:
            node.convert(use_tag=self.options['use_tag'])

        return dict(
            map(
                lambda node: node.to_dict(),
                self._nodes
            )
        )


class Node(TypeChecker):
    FUNCTION_MAPPING = {
        'Ref': Ref,
        # 'Fn::Base64',
        # 'Fn::FindInMap',
        # 'Fn::GetAtt',
        # 'Fn::GetAZs',
        'Fn::ImportValue': ImportValue,
        'Fn::Join': Join,
        # 'Fn::Select',
        'Fn::Sub': Sub,
        # 'Fn::And',
        # 'Fn::Equals',
        # 'Fn::If',
        # 'Fn::Not',
        # 'Fn::Or'
    }

    def __init__(self, key, value, parser, parent):
        self._key = key
        self._value = value
        self._raw_value = value
        self._parser = parser
        self.parent = parent

    def to_dict(self):
        return (
            self._key,
            self._value
        )

    def _get_function_node(self, value):
        return self.FUNCTION_MAPPING[self._get_function_key(value)]

    def convert(self, use_tag=True):
        if self._is_function(self._value):
            # If we're converting joins to subs
            if self._parser.options['join_to_sub'] and (self._get_function_key(self._value) == 'Fn::Join'):
                function__node_class = self.FUNCTION_MAPPING['Fn::Sub']
            else:
                function__node_class = self._get_function_node(self._value)

            _parser_options = self._parser.options.copy()
            _parser_options['parser_class'] = self._parser.__class__

            self._value = function__node_class(self._value, **_parser_options)
            self._value
        else:
            print 'not_function'
        return self



import yaml


class FindInMap(yaml.YAMLObject):
    yaml_tag = u'!FindInMap'

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return '{tag}({value})'.format(tag=self.__class__.__name__, value=self.value)

    @staticmethod
    def representer(dumper, data):
        return dumper.represent_sequence(u'!FindInMap', data.value, flow_style=True)

    @staticmethod
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return FindInMap(value)

if __name__ == 'cfnjsontoyaml.yamlobject.findinmap':
    yaml.add_representer(FindInMap, FindInMap.representer)
    yaml.add_constructor(u'!FindInMap', FindInMap.constructor)



# Builds a Sub function from a join.
import random
import string

import re

class SubBuilder(object):
    FUNCTIONS = [
        'Ref',
        # 'Fn::Base64',
        # 'Fn::FindInMap',
        'Fn::GetAtt',
        # 'Fn::GetAZs',
        'Fn::ImportValue',
        'Fn::Join',
        # 'Fn::Select',
        'Fn::Sub'
        # 'Fn::And',
        # 'Fn::Equals',
        # 'Fn::If',
        # 'Fn::Not',
        # 'Fn::Or'
    ]

    def __init__(self, join_token, values):
        self._join_token = join_token
        self._values = values

        # If we have any values which can't be converted to elements in the sub string, put them in here.
        self._string_substitutions = {}
        self._pattern = None

    def build(self):
        # If this is a join then we will have a list of values to join.

        sub_pattern = []
        substitutions = []

        for value in self._values:
            pattern_key, substitution = self.stringify_value(value)
            if substitution is not None:
                substitutions.append(substitution)
            sub_pattern.append(pattern_key)

        string_pattern = self._join_token.join(sub_pattern)

        return string_pattern, dict(substitutions) if len(substitutions) else None

    def stringify_value(self, value):
        value_type = self._get_value_type(value)
        random_key = ''.join(random.choice('abcdef' + string.digits) for _ in range(8))

        if value_type in ['str', 'unicode']:
            return value, None
        elif value_type == 'Ref':
            return '${' + value['Ref'] + '}', None
        elif value_type == 'Fn::GetAtt':
            return '${' + value['Fn::GetAtt'][0] + '.' + value['Fn::GetAtt'][1] + '}', None
        elif value_type in ['Fn::ImportValue']:
            sub_key = re.sub(r'[^a-z]+', '', value_type.lower()) + '_' + random_key
            return '${' + sub_key + '}', (sub_key, value)
        elif value_type in ['!Ref','!GetAtt']:
            return '${' + value.value + '}', None
        elif value_type.startswith('!'):
            if value_type == '!Sub':
                if not value.is_sequence:
                    return value.value, None
            sub_key = re.sub(r'[^a-z]+', '', value.yaml_tag.lower()) + '_' + random_key
            sub_key = sub_key.encode('ascii', 'ignore')
            return '${' + sub_key + '}', (sub_key, value)
        else:
            return value

    # # 1 - do a check to make sure the tree below can be converted
    #     values_are_stringable = self.validate_value_string_conversion()
    #     # 2 - iterate through each element in the values and convert it to a string.
    #     if values_are_stringable:
    #         converted_values = self.convert_values_to_string()
    #
    # def convert_values_to_string(self):
    #
    #
    # def validate_value_string_conversion(self):
    #     all_values_valid = True
    #     for value in self._values:
    #         value_type = self._get_value_type(value)
    #         if value_type == 'str':
    #             all_values_valid &= True
    #         else:
    #             all_values_valid &= False
    #     return all_values_valid

    def _get_value_type(self, value):
        _type = type(value).__name__
        if _type in ['str', 'unicode']:
            return type(value).__name__
        if (_type == 'dict') and (len(value.keys()) == 1) and (value.keys()[0] in self.FUNCTIONS):
            return value.keys()[0]
        try:
            return value.yaml_tag
        except AttributeError as error:
            raise AttributeError("Could not find 'yaml_tag' on {0}:{1}".format(value, type(value).__name__))



import yaml

from cfnjsontoyaml.mixins import FUNCTION_MAPPING
from cfnjsontoyaml.mixins.type_checker import TypeChecker
from cfnjsontoyaml.parser.subbuilder import SubBuilder


class ConvertToMediary(TypeChecker):
    def __init__(self, template):
        self._template = template
        self._mapping = FUNCTION_MAPPING

    def convert(self):
        return self.walk_dictionary(self._template)

    def walk_list(self, array):
        parsed_list = []
        for value in array:
            value_type = type(value)
            if value_type == dict:
                parsed_list.append(self.walk_dictionary(value))
            elif value_type in [str, unicode]:
                parsed_list.append(value)
            elif value_type == list:
                parsed_list.append(self.walk_list(value))
            else:
                parsed_list.append(value)
        return parsed_list

    def walk_dictionary(self, dictionary, use_literal=False):
        parsed_dictionary = {}
        for key, value in dictionary.items():
            value_type = type(value)
            if value_type == dict:
                if key in self.FUNCTIONS:
                    if key == 'Fn::Join':
                        parsed_value = SubBuilder(*parsed_value).build()
                    else:
                        parsed_value = value
                    parsed_value = self._mapping[key](parsed_value)
                else:
                    parsed_value = value
                parsed_dictionary[key] = self.walk_dictionary(value, use_literal=(key == 'Fn::Base64'))
            elif value_type in [str, unicode]:
                if key in self.FUNCTIONS:
                    if key == 'Fn::Join':
                        parsed_value = SubBuilder(*parsed_value).build()
                    else:
                        parsed_value = value
                    return self._mapping[str(key)](parsed_value)
                else:
                    parsed_dictionary[key] = value
            elif value_type == list:
                parsed_value = self.walk_list(value)
                kwargs = {}
                if key in self.FUNCTIONS:
                    if str(key) == 'Fn::Join':
                        key = 'Fn::Sub'
                        parsed_value = SubBuilder(*parsed_value).build()
                        kwargs['use_literal'] = use_literal
                        return self._mapping[str(key)](parsed_value, **kwargs)
                    else:
                        return self._mapping[str(key)](parsed_value)

                else:
                    parsed_dictionary[key] = value
            else:
                value
        return parsed_dictionary



import yaml


class Ref(yaml.YAMLObject):
    yaml_tag = u'!Ref'

    def __init__(self, value):
        self.value = value

    def __to_sub_string(self):
        return self.value

    def __repr__(self):
        return '{tag}({value})'.format(tag=self.__class__.__name__, value=self.value)

    @staticmethod
    def representer(dumper, data):
        return dumper.represent_scalar(u'!Ref', data.value, style='plain')

    @staticmethod
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return Ref(value)

if __name__ == 'cfnjsontoyaml.yamlobject.ref':
    yaml.add_representer(Ref, Ref.representer)
    yaml.add_constructor(u'!Ref', Ref.constructor)



from cfnjsontoyaml.parser.ref import RefParser


class FunctionParser(object):
    def __init__(self, node):
        self._node = node

    @classmethod
    def factory(cls, _node):
        function_name = cls._get_function_key(_node)
        return {
            'Ref': RefParser
        }[function_name](_node)

    @classmethod
    def _get_function_key(self, node):
        return node.keys()[0]



import yaml


class GetAtt(yaml.YAMLObject):
    yaml_tag = u'!GetAtt'

    def __init__(self, attribute):
        self.value = "{0}.{1}".format(*attribute)

    def __repr__(self):
        return '{tag}({value})'.format(tag=self.__class__.__name__, value=self.value)

    @staticmethod
    def representer(dumper, data):
        return dumper.represent_scalar(u'!GetAtt', data.value, style='double-quoted')

    @staticmethod
    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return GetAtt(value)


if __name__ == 'cfnjsontoyaml.yamlobject.getatt':
    yaml.add_representer(GetAtt, GetAtt.representer)
    yaml.add_constructor(u'!GetAtt', GetAtt.constructor)



from setuptools import setup, find_packages

__version__ = '1.0.3'

setup(
    name='cfnjsontoyaml',
    version=__version__,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    url='https://github.com/drewsonne/cfn-json-to-yaml',
    download_url='https://github.com/drewsonne/cfn-json-to-yaml/archive/v.{0}.zip'.format(__version__),  # I'll explain this in a second
    license='LGPL',
    author='Drew J. Sonne',
    author_email='drew.sonne@gmail.com',
    description='',
    entry_points={
        'console_scripts': [
            'cfn-json-to-yaml=cfnjsontoyaml.__main__:convert'
        ]
    },
    install_requires=[
        'pyyaml',
        'six'
    ],

)



import json, sys, yaml

from cfnjsontoyaml.convertor import ConvertToMediary
from cfnjsontoyaml.yamlobject.unicode import CleanDumper


def print_to_string(document):
    return yaml.dump(
        document,
        Dumper=CleanDumper,
        encoding='utf-8',
        allow_unicode=True,
        explicit_start=True
    )


def convert(args=sys.argv[1:]):
    if args:
        with open(args[0]) as fp:
            json_document = fp.read()
    else:
        json_document = []
        for line in sys.stdin:
            json_document.append(line)

        json_document = "\n".join(json_document)

    document = json.loads(json_document)

    print print_to_string(ConvertToMediary(document).convert())


if __name__ == '__main__':
    convert()
